/**
 * Copyright (C) 2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.genclass.map;

import org.objectweb.speedo.genclass.GenClassElement;
import org.objectweb.speedo.genclass.GenClassAccessor;
import org.objectweb.speedo.mim.api.SpeedoProxy;
import org.objectweb.speedo.pm.api.ProxyManager;
import org.objectweb.speedo.genclass.api.SpeedoGenClassProxy;
import org.objectweb.speedo.mim.api.SpeedoAccessor;
import org.objectweb.jorm.api.PIndexedElem;
import org.objectweb.jorm.api.PExceptionIO;
import org.objectweb.jorm.naming.api.PName;

import java.util.Map;
import java.util.Iterator;
import java.util.Set;
import java.util.Collection;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;

import javax.jdo.PersistenceManager;

public class MapAccessor
        extends GenClassAccessor
        implements Map {

    /**
     * A map containing the indexed elements of the genclass. It associates each
     * element to its index. This map does NOT contain the deleted elements.
     */
    protected Map map = null;

    /**
     * Instanciates and initializes a new map with an initial size.
     */
    public MapAccessor(SpeedoGenClassProxy jdoProxy) {
        super(jdoProxy);
        this.elements = new ArrayList();
        this.map = (Map) jdoProxy.createGenClass();
    }

	public void jdoDeletePersistent(ProxyManager pm) {
		clear();
	}

	public void setGCValue(Object o) {
		clear();
		putAll(((Map) o));
	}

	public void jdoLoadFieldsFromAccessor(SpeedoAccessor sa) {
		MapAccessor ma = (MapAccessor) sa;
		map.clear();
		map.putAll(ma.map);
		inner = ma.getInner();
		supportDelta = ma.supportDelta;
		elements.clear();
		elements.addAll(ma.elements);
		jdoSetStatus(ma.jdoGetStatus());
	}

	public boolean speedoAdd(Object elemToAdd, Object hints) {
		put(hints, elemToAdd, false);
		return true;
	}

	public boolean speedoRemove(Object elemToRemove, Object hints) {
		return remove(hints, false) != null;
	}

	public Object put(Object key, Object value, boolean withCoherence) {
		// Put it in "elements" and "map"
		GenClassElement element = (GenClassElement) createPIndexedElem();
		if (value instanceof SpeedoProxy) {
			jdoProxy.jdoMakePersistent(null, (SpeedoProxy) value);
		}
		element.setIndex(key);
		element.setElement(value);
		elements.add(element); // TODO et si element est deja dans elements? s'il est marque deleted?
		Object oldValue = get(key);
		map.put(key, element);
		if (withCoherence) {
			Object _o = value;
			if (_o instanceof PName) {
				_o = element.getElement(jdoProxy.jdoGetPersistenceManager());
			}
			if (_o instanceof SpeedoProxy) {
				jdoProxy.fireSpeedoElementAdded(_o);
			}
		}
		return oldValue;
	}

	public Object remove(Object key, boolean withCoherence) {
		GenClassElement gcelem = (GenClassElement) map.get(key);
		if (gcelem == null) {
			return null;
		}
		gcelem.setStatus(PIndexedElem.ELEM_DELETED);
		Object value = get(key);
		map.remove(key);
		if (withCoherence) {
			Object _o = gcelem.getElement(jdoProxy.jdoGetPersistenceManager());
			if (_o != null) {
				jdoProxy.fireSpeedoElementRemoved(_o);
			}
		}
		return value;
	}

    // ------------------------------------------------------------------------
    // IMPLEMENTATION OF THE Map INTERFACE
    // ------------------------------------------------------------------------

    public int size() {
        return map.size();
    }

    public boolean isEmpty() {
        return map.isEmpty();
    }

    public boolean containsKey(Object key) {
        return map.containsKey(key);
    }

    public boolean containsValue(Object value) {
        // TODO: support the containsValue method on Map
        return false;
    }

    public Object get(Object key) {
        Object value = map.get(key);
        if (value == null) {
            return null;
        } else {
            return ((MapElem) value).getElement(
                    jdoProxy.jdoGetPersistenceManager());
        }
    }

    public Object put(Object key, Object value) {
		return put(key, value, true);
	}

    public Object remove(Object key) {
		return remove(key, true);
    }

    public void putAll(Map t) {
        Iterator i = t.entrySet().iterator();
        while (i.hasNext()) {
            Map.Entry entry = (Map.Entry) i.next();
            put(entry.getKey(), entry.getValue());
        }
    }

    public void clear() {
        Iterator i = elements.iterator();
		PersistenceManager pm = null;
        while (i.hasNext()) {
            GenClassElement gcelem = (GenClassElement) i.next();
            gcelem.setStatus(PIndexedElem.ELEM_DELETED);
			if (pm == null) {
				pm = jdoProxy.jdoGetPersistenceManager();
			}
			Object _o = gcelem.getElement(pm);
			if (_o != null) {
				jdoProxy.fireSpeedoElementRemoved(_o);
			}
        }
        map.clear();
		supportDelta = false;
    }

    public Set keySet() {
	    return Collections.unmodifiableSet(map.keySet());
    }

    public Collection values() {
	    ArrayList res = new ArrayList(elements.size());
	    for(int i=(elements.size()-1); i>=0; i--) {
		    MapElem me = (MapElem) elements.get(i);
		    if (me.getElemStatus() != PIndexedElem.ELEM_DELETED) {
			    res.add(me.getElement());
		    }
 	    }
        return Collections.unmodifiableCollection(res);
    }

    public Set entrySet() {
	    Set res = new HashSet(elements.size());
	    for(int i=(elements.size()-1); i>=0; i--) {
		    MapElem me = (MapElem) elements.get(i);
		    if (me.getElemStatus() != PIndexedElem.ELEM_DELETED) {
			    res.add(new MyMapEntry(me));
		    }
 	    }
	    return Collections.unmodifiableSet(res);
    }

	private class MyMapEntry implements Map.Entry {
		MapElem me;

		public MyMapEntry(MapElem me) {
			this.me = me;
		}

		public Object getKey() {
			return me.getIndex();
		}

		public Object getValue() {
			return me.getElement();
		}

		public Object setValue(Object value) {
			Object res = me.getElement();
			me.setStatus(PIndexedElem.ELEM_MODIFIED);
			me.setElement(value);
			return res;
		}
	}

    // ------------------------------------------------------------------------
    // IMPLEMENTATION OF THE PGenClassAccessor INTERFACE
    // ------------------------------------------------------------------------
    public PIndexedElem createPIndexedElem(GenClassAccessor gca) {
        return new MapElem(gca);
    }

    public void paAdd(PIndexedElem elem, Object conn) throws PExceptionIO {
        GenClassElement gcelem = (GenClassElement) elem;
        // the elem is read from the DS, set it to unmodified.
        gcelem.setStatus(PIndexedElem.ELEM_UNMODIFIED);
        elements.add(gcelem);
        map.put(gcelem.getIndex(), gcelem);
    }

    public int paGetNbElem() {
        return size();
    }

    public Iterator paIterator() {
        return elements.iterator();
    }

    public void paSetNbElem(int nbelem) {
	    if (nbelem == -1) {
            elements = new ArrayList();
	    } else {
            elements = new ArrayList(nbelem);
	    }
        map = (Map) jdoProxy.createGenClass();
    }
}
