/**
 * Copyright (C) 2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.pm.lib;

import org.objectweb.fractal.adl.Launcher;
import org.objectweb.fractal.adl.Parser;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.Interface;
import org.objectweb.fractal.api.factory.GenericFactory;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.fractal.api.control.IllegalContentException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.util.Fractal;
import org.objectweb.perseus.pool.api.PoolMatchFactory;
import org.objectweb.perseus.pool.api.PoolException;
import org.objectweb.perseus.persistence.api.ConnectionHolderFactory;
import org.objectweb.perseus.persistence.api.PersistenceException;
import org.objectweb.speedo.pm.api.ProxyManager;
import org.objectweb.speedo.workingset.api.Transaction;
import org.objectweb.speedo.workingset.lib.SpeedoTransaction;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import java.util.HashMap;
import java.util.Map;

/**
 *
 * @author S.Chassande-Barrioz
 */
public class SpeedoProxyManagerInstanciator
		implements PoolMatchFactory, BindingController, LifeCycleController {


	public final static String CONNECTION_HOLDER_FACTORY_BINDING = "connection-holder-factory";
	public final static String COMPONENT_BINDING = "component";
	public final static String PM_TEMPLATE = "org.objectweb.speedo.pm.lib.SpeedoProxyManager";
	public final static String T_TEMPLATE = "org.objectweb.speedo.workingset.lib.SpeedoTransaction";

	private ConnectionHolderFactory chf;
	private Component thisC = null;
	private Component pmT;
	private Component tT;

	private boolean started = false;

	private Logger logger;
	Map components = new HashMap(5);
	Component speedoComponent = null;

	void clearComponentRef() {
		components.clear();
	}

	void fetchComponentRef() throws Exception {
		speedoComponent = Fractal.getSuperController(thisC).getFcSuperComponents()[0];
		Component[] children = Fractal.getContentController(speedoComponent)
				.getFcSubComponents();
		for(int i=0; i<children.length; i++) {
			String name = Fractal.getNameController(children[i]).getFcName();
			if ("proxy-manager-factory".equals(name)) {
				components.put(name, children[i].getFcInterface("proxy-manager-factory"));
			} else if ("mapper".equals(name)) {
				components.put("mapper", children[i].getFcInterface("mapper"));
				components.put("jorm-factory", children[i].getFcInterface("jorm-factory"));
				components.put("pname-coder", children[i].getFcInterface("pname-coder"));
			} else if ("query-manager".equals(name)) {
				components.put(name, children[i].getFcInterface("query-manager"));
			} else if ("tpm".equals(name)) {
				components.put(name, children[i].getFcInterface("transactional-persistence-manager"));
			}
		}
		Parser parser = Launcher.getBootstrapParser();
		pmT = parser.loadTemplate(PM_TEMPLATE, true);
		tT = parser.loadTemplate(T_TEMPLATE, true);
	}

	// IMPLEMENTATION OF THE UserBindingController INTERFACE //
	//-------------------------------------------------------//

	public String getFcState() {
		return started ? STARTED : STOPPED;
	}

	public void startFc() throws IllegalLifeCycleException {
		started = true;
		try {
			fetchComponentRef();
		} catch (Exception e) {
			if (logger != null) {
				logger.log(BasicLevel.ERROR, "Impossible to fetch inner speedo components", e);
			}
			throw new IllegalLifeCycleException(e.getMessage());
		}
	}

	public void stopFc() throws IllegalLifeCycleException {
		started = false;
		clearComponentRef();
	}


	// IMPLEMENTATION OF THE UserBindingController INTERFACE //
	//-------------------------------------------------------//

    public String[] listFc() {
        return new String[] { CONNECTION_HOLDER_FACTORY_BINDING };
    }

	public Object lookupFc(String c) {
		if (CONNECTION_HOLDER_FACTORY_BINDING.equals(c)) {
			return chf;
		} else {
			return null;
		}
	}

	public void bindFc(String c, Object s) {
		if ("logger".equals(c)) {
			logger = (Logger) s;
		} else if (CONNECTION_HOLDER_FACTORY_BINDING.equals(c)) {
			chf = (ConnectionHolderFactory) s;
		} else if (COMPONENT_BINDING.equals(c)) {
			thisC = (Component) s;
		}
	}

	public void unbindFc(String c) {
		if (CONNECTION_HOLDER_FACTORY_BINDING.equals(c)) {
			chf = null;
		}
	}


	// IMPLEMENTATION OF THE PoolMathcFactory INTERFACE //
	//--------------------------------------------------//

	/**
	 * Creates a new <code>PoolResource</code>.
	 * This methos is invoked by the owned <code>Pool</code>.
	 * A new SpeedoProxyManager and a new SpeedoTransaction component are
	 * created, and added into the Speedo composite.
	 *
	 * @param o params used to build a new PoolResource
	 * @return the build <code>SpeedoProxyManager</code>
	 */
	public Object createResource(Object o) throws PoolException {
		try {
			//instanciate the ProxyManager
			Component pmC = Fractal.getFactory(pmT).newFcInstance();
			ProxyManager pm = (ProxyManager) pmC.getFcInterface("proxy-manager");
			Fractal.getNameController(pmC).setFcName("proxy-manager");
			BindingController pmBC = Fractal.getBindingController(pmC);

			//instanciate the Transaction
			Component tC = Fractal.getFactory(tT).newFcInstance();
			Fractal.getNameController(tC).setFcName("transaction");
			Transaction t = (Transaction) tC.getFcInterface("transaction");
			BindingController tBC = Fractal.getBindingController(tC);

			//Add the new components into the Speedo composite
			Fractal.getContentController(speedoComponent).addFcSubComponent(pmC);
			Fractal.getContentController(speedoComponent).addFcSubComponent(tC);

			//add bindings on the proxy manager
			pmBC.bindFc(SpeedoProxyManager.JORM_FACTORY_BINDING,
					components.get("jorm-factory"));
			pmBC.bindFc(SpeedoProxyManager.PNAME_CODER_BINDING,
					components.get("pname-coder"));
			pmBC.bindFc(SpeedoProxyManager.PROXY_MANAGER_FACTORY_BINDING,
					components.get("proxy-manager-factory"));
			pmBC.bindFc(SpeedoProxyManager.QUERY_MANAGER_BINDING,
					components.get("query-manager"));
			pmBC.bindFc(SpeedoProxyManager.TRANSACTIONAL_PERSISTENCE_MANAGER_BINDING,
					components.get("tpm"));
			pmBC.bindFc(SpeedoProxyManager.TRANSACTION_BINDING, t);

			//add bindings on the transaction
			tBC.bindFc(SpeedoTransaction.PROXY_MANAGER_BINDING, pm);
			tBC.bindFc(SpeedoTransaction.TRANSACTIONAL_PERSISTENCE_MANAGER_BINDING,
					components.get("tpm"));
			tBC.bindFc(SpeedoTransaction.MAPPER_BINDING,
					components.get("mapper"));

			//Add a ConenctionHolder to the transaction
			t.setConnectionHolder(chf.createConnectionHolder());

			//start components
			Fractal.getLifeCycleController(pmC).startFc();
			Fractal.getLifeCycleController(tC).startFc();

			return pm;
		} catch (Exception e) {
			PoolException pe = new PoolException(
			    "Error during the allocation of a new ProxyManager and its WorkingSet", e);
			logger.log(BasicLevel.ERROR, pe.getMessage(), pe);
			throw pe;
		}
	}

	public boolean matchResource(Object resource, Object o) {
		return true;
	}

	/**
	 * It removes the ProxyManager and the Transaction components from the
	 * Speedo composite.
	 * @param resource is the ProxyManager to destroy.
	 */
	public void destroyResource(Object resource) {
		try {
			//fetch the PM component
			Component pmC = ((Interface) resource).getFcItfOwner();
			BindingController pmBC = Fractal.getBindingController(pmC);

			//fetch the Tx component
			Component tC = ((Interface)
					pmBC.lookupFc(SpeedoProxyManager.TRANSACTION_BINDING))
					.getFcItfOwner();
			BindingController tBC = Fractal.getBindingController(tC);

			//Stop compoenents
			Fractal.getLifeCycleController(pmC).stopFc();
			Fractal.getLifeCycleController(tC).stopFc();

			//remove bindings of the PM component
			pmBC.unbindFc(SpeedoProxyManager.JORM_FACTORY_BINDING);
			pmBC.unbindFc(SpeedoProxyManager.PNAME_CODER_BINDING);
			pmBC.unbindFc(SpeedoProxyManager.PROXY_MANAGER_FACTORY_BINDING);
			pmBC.unbindFc(SpeedoProxyManager.QUERY_MANAGER_BINDING);
			pmBC.unbindFc(SpeedoProxyManager.TRANSACTIONAL_PERSISTENCE_MANAGER_BINDING);
			pmBC.unbindFc(SpeedoProxyManager.TRANSACTION_BINDING);

			//remove bindings of the Tx component
			tBC.unbindFc(SpeedoTransaction.PROXY_MANAGER_BINDING);
			tBC.unbindFc(SpeedoTransaction.TRANSACTIONAL_PERSISTENCE_MANAGER_BINDING);
			tBC.unbindFc(SpeedoTransaction.MAPPER_BINDING);

			//remove both components from the composite
			Fractal.getContentController(speedoComponent).addFcSubComponent(pmC);
			Fractal.getContentController(speedoComponent).addFcSubComponent(tC);
		} catch (Exception e) {
			if (logger != null) {
				logger.log(BasicLevel.WARN, "Error during the destroying of a ProxyManager", e);
			}
		}
	}

}
