/**
 * Copyright (C) 2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.query.lib;

import org.objectweb.speedo.query.api.QueryDefinition;

import javax.jdo.Extent;
import java.util.List;
import java.util.Collection;

/**
 *
 * @author S.Chassande-Barrioz
 */
public class QueryDefinitionImpl implements QueryDefinition {
	/**
	 * IgnoreCache option.
	 * The ignoreCache option setting specifies whether the query should execute
	 * entirely in the back end, instead of in the cache.  If this flag is set
	 * to true, an implementation might be able to optimize the query
	 * execution by ignoring changed values in the cache.  For optimistic
	 * transactions, this can dramatically improve query response times.
	 */
	protected boolean ignoreCache = true;

	/**
	 * Candidate classes or extent for the query.
	 */
	protected Collection candidateInstances = null;
	protected Extent extentClass = null;

	/**
	 * Class of candadate classes.
	 */
	protected Class candidateClass = null;

	/**
	 * Query filter.
	 */
	protected String filter = "true";

	/**
	 * Query parameters and variables declaration
	 */
	protected String parameters = null;
	protected String variables = null;

	/**
	 * the import statements is transformed into a simple vector
	 */
	protected List importStatements = null;
	protected List order = null;

	public QueryDefinitionImpl() {
	}

	public QueryDefinitionImpl(QueryDefinition qd) {
		this.candidateClass = qd.getCandidateClass();
		this.candidateInstances = qd.getCollection();
		this.extentClass = qd.getExtent();
		this.filter = qd.getFilter();
		this.order = qd.getOrder();
		this.variables = qd.getVariables();
		this.parameters = qd.getVariables();
		this.importStatements = qd.getImportStatements();
		this.ignoreCache = qd.isIgnoreCache();
	}

	public void defineWith(QueryDefinition qd) {
	    if (qd == null) {
		    this.candidateClass = null;
		    this.candidateInstances = null;
		    this.extentClass = null;
		    this.filter = null;
		    this.order = null;
		    this.variables = null;
		    this.parameters = null;
		    this.importStatements = null;
		    this.ignoreCache = true;
	    } else {
		    this.candidateClass = qd.getCandidateClass();
		    this.candidateInstances = qd.getCollection();
		    this.extentClass = qd.getExtent();
		    this.filter = qd.getFilter();
		    this.order = qd.getOrder();
		    this.variables = qd.getVariables();
		    this.parameters = qd.getVariables();
		    this.importStatements = qd.getImportStatements();
		    this.ignoreCache = qd.isIgnoreCache();
	    }
	}


	public Class getCandidateClass() {
		return candidateClass;
	}

	public boolean isCollection() {
		return candidateInstances != null;
	}

	public Collection getCollection() {
		return candidateInstances;
	}

	public String getParameters() {
		return parameters;
	}

	public String getVariables() {
		return variables;
	}

	public String getFilter() {
		return filter;
	}

	public Extent getExtent() {
		return extentClass;
	}

	public List getImportStatements() {
		return importStatements;
	}

	public boolean isIgnoreCache() {
		return ignoreCache;
	}

	public List getOrder() {
		return order;
	}


	public String qdToString(boolean oneLine) {
		StringBuffer sb = new StringBuffer();
		String sep = "";
		final String sepNext = (oneLine ? ", " : "\n\t-");
		if (candidateClass!=null) {
			sb.append(sep);
			sb.append("candidateClass=");
			sb.append(candidateClass.getName());
			sep = sepNext;
		}
		if (parameters!=null && parameters.length() >0) {
			sb.append(sep);
			sb.append("parameters=\"");
			sb.append(parameters);
			sb.append("\"");
			sep = sepNext;
		}
		if (variables!=null && variables.length() >0) {
			sb.append(sep);
			sb.append("variables=\"");
			sb.append(variables);
			sb.append("\"");
			sep = sepNext;
		}
		if (filter!=null && filter.length() >0) {
			sb.append(sep);
			sb.append("filter=\"");
			sb.append(filter);
			sb.append("\"");
			sep = sepNext;
		}
		if (order!=null && order.size() >0) {
			sb.append(sep);
			sb.append("order=");
			sb.append(order);
			sep = sepNext;
		}
		sb.append(sep);
		sb.append("ignoreCache=");
		sb.append(ignoreCache);
		sep = sepNext;
		return sb.toString();
	}

	public int hashCode() {
		return candidateClass.hashCode();
	}

	public boolean equals(Object obj) {
		if (obj == null || !(obj instanceof QueryDefinition)) {
			return false;
		}
		QueryDefinition qd = (QueryDefinition) obj;
		return (candidateClass==null
				? qd.getCandidateClass()==null
				: candidateClass.equals(qd.getCandidateClass()))
				&& (parameters == null
				? qd.getParameters() == null
				: parameters.equals(qd.getParameters()))
				&& (variables == null
				? qd.getVariables() == null
				: variables.equals(qd.getVariables()))
				&& (filter == null
				? qd.getFilter() == null
				: filter.equals(qd.getFilter()))
				&& (order == null
				? qd.getOrder() == null
				: order.equals(qd.getOrder()))
				&& (qd.isCollection()
					? qd.getCollection().equals(candidateInstances) : true)
				&& (qd.isIgnoreCache() == ignoreCache)
				;
	}

}
