/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 0.1
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: X. Spengler
 *
 */

package org.objectweb.speedo.query.parser;

import org.objectweb.medor.api.Field;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.api.TypingException;
import org.objectweb.medor.expression.lib.And;
import org.objectweb.medor.expression.lib.BasicOperand;
import org.objectweb.medor.expression.lib.Bitwize;
import org.objectweb.medor.expression.lib.ConditionalAnd;
import org.objectweb.medor.expression.lib.ConditionalOr;
import org.objectweb.medor.expression.lib.DivideBy;
import org.objectweb.medor.expression.lib.Equal;
import org.objectweb.medor.expression.lib.Greater;
import org.objectweb.medor.expression.lib.GreaterEqual;
import org.objectweb.medor.expression.lib.Length;
import org.objectweb.medor.expression.lib.Like;
import org.objectweb.medor.expression.lib.Lower;
import org.objectweb.medor.expression.lib.LowerEqual;
import org.objectweb.medor.expression.lib.Minus;
import org.objectweb.medor.expression.lib.Mult;
import org.objectweb.medor.expression.lib.Not;
import org.objectweb.medor.expression.lib.NotEqual;
import org.objectweb.medor.expression.lib.Or;
import org.objectweb.medor.expression.lib.Plus;
import org.objectweb.medor.expression.lib.StringLower;
import org.objectweb.medor.expression.lib.StringUpper;
import org.objectweb.medor.expression.lib.UMinus;
import org.objectweb.medor.expression.lib.Substring;
import org.objectweb.medor.filter.lib.BasicFieldOperand;
import org.objectweb.medor.filter.lib.ExpressionPrinter;
import org.objectweb.medor.filter.lib.MemberOf;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.jorm.lib.NavigatorNodeFactory;
import org.objectweb.util.monolog.api.BasicLevel;

import javax.jdo.JDOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Stack;

/**
 * Implementation of a visitor that creates the filter
 */
public class SpeedoQLQueryFilterVisitor extends SpeedoQLAbstractVisitor {

    private final static String CONTAINS_PATH_SET = "Next element is the contain path set";
    private final static String REMOVER = "Remover";


    /**
     * qf is the transformation of the filter into a MEDOR expression
     */
    private Expression qf = null;

    /**
     * qt is built by the SpeedoQLVariableVisitor and is just used here
     */
    private QueryTree qt = null;

    /**
     * speedoql is the result of the filter parsing
     */
    private ASTSpeedoQL speedoql = null;

    private String tab ="";

    private boolean debug = false;

    /**
     * Constructor
     * @param qtree a QueryTree built with the list of variables, parameters and name.
     * @param speedoql root of the lexical tree of the query
     */
    public SpeedoQLQueryFilterVisitor(QueryTree qtree,
                                      ASTSpeedoQL speedoql) {
        qt = qtree;
        this.speedoql = speedoql;
    }

    /**
     * the visit starts here, please before sets a list of variables.
     * @throws java.lang.Exception
     */
    public void startVisiting() throws Exception {
        debug = logger.isLoggable(BasicLevel.DEBUG);
        try {
            visit(speedoql);
        } catch (SpeedoQLAbstractVisitor.VisitorException e) {
            throw e.getNestedException();
        }
    }

    /**
     * get the query filter that was built from visiting the syntaxic tree
     */
    public Expression getQueryFilter() {
        return qf;
    }

    public Object visit(ASTPrimary node, Object data) {
		Stack stack = (Stack) data;
		boolean hasMethod = node.value != null;
		if (hasMethod) {
			stack.push(node.value);
		}
        visit((SimpleNode) node, data);
		if (hasMethod) {
			Expression e2 = (Expression) stack.pop();
			Expression e1 = (Expression) stack.pop();
			String methodName = (String) stack.pop();
			if (methodName.startsWith("null.")) {
				methodName = methodName.substring("null.".length());
			}
			stack.push(e1);
			stack.push(methodName);
			e1 = treatMethodOperator(methodName, stack, e2);
			if (e1 != null) {
				stack.push(e1);
			}
		}
        return null;
    }

    public Object visit(ASTSpeedoPrimary node, Object data) {
        visit((SimpleNode) node, data);
        Stack stack = (Stack) data;
        while(stack.size() >0 && REMOVER.equals(stack.peek())) {
            stack.pop();
        }
        qf = (Expression) stack.pop();
        return null;
    }

    /**
     *
     */
    public Object visit(ASTRelationalExpression node, Object data) {
        logger.log(BasicLevel.DEBUG, tab + "Visit RelationalExpression: " + node);
        tab += '\t';
        visit((SimpleNode) node, data);
        tab = tab.substring(1);
        Stack stack = (Stack) data;
        if (stack.size() > 0) {
            if (CONTAINS_PATH_SET.equals(stack.peek())) {
                if (debug) {
                    logger.log(BasicLevel.ERROR, "remove the element of the stack");
                }
            } else {
                if (debug) {
                    logger.log(BasicLevel.DEBUG, "manage relational expression: "
                        + "(children: " + node.jjtGetNumChildren()
                        + ", stack: " + stack.size()
                        + ", peek:" + stack.peek() + ")");
                }
                for (int i = 0; (i < (node.jjtGetNumChildren() - 1) && stack.size() > 0); i++) {
                    int op = ((Integer) node.ops.get(node.jjtGetNumChildren() - 2 - i)).intValue();
                    if (usedInRelationalExpresssion(op)) {
                        Object child2 = stack.pop();
                        Object child1 = stack.pop();

                        if (debug) {
                            logger.log(BasicLevel.DEBUG, "pop child1: " + child1);
                            logger.log(BasicLevel.DEBUG, "pop child2: " + child2);
                        }
                        if (REMOVER.equals(child1)) {
                            if (REMOVER.equals(child2)) {
                                //nothin
                            } else {
                                stack.push(child2);
                                if (debug) {
                                    logger.log(BasicLevel.DEBUG, "push(" + child2 + ")");
                                }
                            }
                        } else {
                            if (REMOVER.equals(child2)) {
                                stack.push(child1);
                                if (debug) {
                                    logger.log(BasicLevel.DEBUG, "push(" + child1 + ")");
                                }
                            } else {
                                Expression ret = null;
                                switch (op) {
                                case SpeedoQLConstants.OR:
                                    ret = new ConditionalOr((Expression) child1, (Expression) child2);
                                    break;
                                case SpeedoQLConstants.AND:
                                    ret = new ConditionalAnd((Expression) child1, (Expression) child2);
                                    break;
                                case SpeedoQLConstants.BITWISEOR:
                                    ret = new Or((Expression) child1, (Expression) child2);
                                    break;
                                case SpeedoQLConstants.BITWISEXOR:
                                    ret = null;
                                    break;
                                case SpeedoQLConstants.BITWISEAND:
                                    ret = new And((Expression) child1, (Expression) child2);
                                    break;
                                case SpeedoQLConstants.EQ:
                                    ret = new Equal((Expression) child1, (Expression) child2);
                                    break;
                                case SpeedoQLConstants.NEQ:
                                    ret = new NotEqual((Expression) child1, (Expression) child2);
                                    break;
                                case SpeedoQLConstants.LT:
                                    ret = new Lower((Expression) child1, (Expression) child2);
                                    break;
                                case SpeedoQLConstants.GT:
                                    ret = new Greater((Expression) child1, (Expression) child2);
                                    break;
                                case SpeedoQLConstants.GTE:
                                    ret = new GreaterEqual((Expression) child1, (Expression) child2);
                                    break;
                                case SpeedoQLConstants.LTE:
                                    ret = new LowerEqual((Expression) child1, (Expression) child2);
                                    break;
                                }
                                if (debug) {
                                    logger.log(BasicLevel.DEBUG, "push(" + ret + ")");
                                }
                                stack.push(ret);
                            }
                        }
                    }
                }
            }
        }
        logger.log(BasicLevel.DEBUG, "children:" + node.jjtGetNumChildren()
                + " / stack: " + stack.size());
        logger.log(BasicLevel.DEBUG, tab + "End of Visit RelationalExpression: " + node);
        return null;
    }

    private boolean usedInRelationalExpresssion(int op) {
        switch (op) {
        case SpeedoQLConstants.OR:
        case SpeedoQLConstants.AND:
        case SpeedoQLConstants.BITWISEOR:
        case SpeedoQLConstants.BITWISEXOR:
        case SpeedoQLConstants.BITWISEAND:
        case SpeedoQLConstants.EQ:
        case SpeedoQLConstants.NEQ:
        case SpeedoQLConstants.LT:
        case SpeedoQLConstants.GT:
        case SpeedoQLConstants.GTE:
        case SpeedoQLConstants.LTE:
            logger.log(BasicLevel.DEBUG, "node useful");
            return true;
        default:
            logger.log(BasicLevel.DEBUG, "node useless");
            return false;
        }
    }

    public Object visit(ASTAdditiveExpression node, Object data) {
        logger.log(BasicLevel.DEBUG, tab + "Visit AdditiveExpression: " + node);
        tab += '\t';
        visit((SimpleNode) node, data);
        tab = tab.substring(1);
        Stack stack = (Stack) data;
        if (stack.size() > 0
                && !CONTAINS_PATH_SET.equals(stack.peek())
                && !REMOVER.equals(stack.peek())) {
            Expression ret = (Expression) stack.pop();
            for (int i = 0; i < node.jjtGetNumChildren() - 1; i++) {
                logger.log(BasicLevel.DEBUG, "Visit ConditionalExpression... children...[" + i + "]");

                switch (((Integer) node.ops.get(node.jjtGetNumChildren() - 2 - i)).intValue()) {

                case SpeedoQLConstants.PLUS:
                    ret = new Plus((Expression) stack.pop(), ret);
                    break;
                case SpeedoQLConstants.MINUS:
                    ret = new Minus((Expression) stack.pop(), ret);
                    break;
                case SpeedoQLConstants.MULT:
                    ret = new Mult((Expression) stack.pop(), ret);
                    break;
                case SpeedoQLConstants.DIV:
                    ret = new DivideBy((Expression) stack.pop(), ret);
                    break;
                default:

                }
            }
            ((Stack) data).push(ret);
        }
        logger.log(BasicLevel.DEBUG, tab + "End of Visit AdditiveExpression: " + node);
        return null;
    }

    public Object visit(ASTUnaryExpression node, Object data) {
        logger.log(BasicLevel.DEBUG, tab + "Visit UnaryExpression" + node);
        tab += '\t';
        visit((SimpleNode) node, data);
        tab = tab.substring(1);
        Stack stack = (Stack) data;
        if (stack.size() > 0) {
            logger.log(BasicLevel.DEBUG, "stack.peek:" + stack.peek());
        }
        if (stack.size() > 0
                && !CONTAINS_PATH_SET.equals(stack.peek())
                && !REMOVER.equals(stack.peek())) {
            Object o = stack.pop();
            if (node.ops.size() > 0) {
                switch (((Integer) node.ops.get(0)).intValue()) {
                case SpeedoQLConstants.MINUS:
                    o = new UMinus((Expression) o);
                    break;
                case SpeedoQLConstants.BITWISECOMPL:
                    o = new Bitwize((Expression) o);
                    break;
                case SpeedoQLConstants.NOT:
                    logger.log(BasicLevel.DEBUG, "NOT(" + o + "): " + node);
                    o = new Not((Expression) o);
                    break;
                }
            }
            ((Stack) data).push(o);
        }
        logger.log(BasicLevel.DEBUG, tab + "End of Visit UnaryExpression: " + node);
        return null;
    }

    // to be done
    public Object visit(ASTCastExpression node, Object data) {
        ((Stack) data).push(node);
        return null;
    }

    /**
     * 4 cases to manage: (en cours par equipe MEDOR)
     * - Collection.contains(Object o)
     * - Collection.isEmpty()
     * - String.startsWith(String s)
     * - String.endsWith(String s)
     */
    public Object visit(ASTArgumentList node, Object data) {
        visit((SimpleNode) node, data);
        return null;
    }

    public Object visit(ASTLiteral node, Object data) {
		Stack stack = (Stack) data;
		Expression e = null;
        if (node.value instanceof Integer) {
            e = new BasicOperand(((Integer) node.value).intValue());
        } else if (node.value instanceof Float) {
            e = new BasicOperand(((Float) node.value).floatValue());
        } else if (node.value instanceof Character) {
            e = new BasicOperand(((Character) node.value).charValue());
        } else if (node.value instanceof String) {
            String s = (String) (node.value);
            s = s.substring(1, s.length()-1);
            e = new BasicOperand(s);
        } else if (node.value instanceof Boolean) {
            e = new BasicOperand(((Boolean) node.value).booleanValue());
		}
		if (stack.size() > 0) {
			Object top = stack.peek();
			if (CONTAINS_PATH_SET.equals(top)) {
//TODO: Support the path.contains(path) operator
				logger.log(BasicLevel.ERROR, "The path.contains(path) operator is not yet implemented");
				return null;
			} else if (top instanceof String) {
				e = treatMethodOperator((String) top, stack, e);
			}
		}
		if (e != null) {
			logger.log(BasicLevel.DEBUG, "push(" + ExpressionPrinter.e2str(e) + ")");
			stack.push(e);
		}
        return null;
    }

	private Expression treatMethodOperator(String opName,
									 Stack stack,
									 Expression e) {
		int op = isMethodOperator(opName);
		if (op == -1) {
			return e;
		}
		switch (op) {
		case STARTS_WITH_OPERATOR:
			{
				logger.log(BasicLevel.DEBUG, "pop: " + stack.pop());
				Object oe = stack.pop();
				logger.log(BasicLevel.DEBUG, "pop expression: " + oe);
				Expression str = (Expression) oe;
				try {
					e = new BasicOperand(((BasicOperand) e).getString() + "%");
				} catch (TypingException e1) {
					throw new JDOException(
						"Bad parameter type for the 'startsWith' method", e1);
				}
				e = new Like(str, e);
				break;
			}
		case ENDS_WITH_OPERATOR:
			{
				logger.log(BasicLevel.DEBUG, "pop: " + stack.pop());
				Object oe = stack.pop();
				logger.log(BasicLevel.DEBUG, "pop expression: " + oe);
				Expression str = (Expression) oe;
				try {
					e = new BasicOperand("%" + ((BasicOperand) e).getString());
				} catch (TypingException e1) {
					throw new JDOException(
						"Bad parameter type for the 'startsWith' method", e1);
				}
				e = new Like(str, e);
				break;
			}
		case EQUALS_OPERATOR:
//TODO: Support the equals operator
			break;
		case MATCHES_OPERATOR:
			{
				logger.log(BasicLevel.DEBUG, "pop: " + stack.pop());
				Object oe = stack.pop();
				logger.log(BasicLevel.DEBUG, "pop expression: " + oe);
				Expression str = (Expression) oe;
//TODO : convert the java.lang.regex pattern to MEDOR/SQL pattern
				e = new Like(str, e);
				break;
			}
		case SUBSTRING_OPERATOR:
			{
				logger.log(BasicLevel.DEBUG, "pop: " + stack.pop());
				Object o = stack.pop();
				if (o == STR_OPERAND_SUBSTRING) {
					logger.log(BasicLevel.DEBUG, "push(" + ExpressionPrinter.e2str(e) + ")");
					stack.push(e);
					logger.log(BasicLevel.DEBUG, "push(BEGIN_OPERAND_SUBSTRING)");
					stack.push(BEGIN_OPERAND_SUBSTRING);
					logger.log(BasicLevel.DEBUG, "push(" + opName + ")");
					stack.push(opName);
					return null;
				} else if (o == BEGIN_OPERAND_SUBSTRING) {
					Expression begin = (Expression) stack.pop();
					Expression str = (Expression) stack.pop();
					e = new Substring(str, begin, e);
				} else {
					throw new JDOException("Protocol error: " + o);
				}
				break;
			}
		}
		return e;
	}

    public Object visit(ASTType node, Object data) {
        logger.log(BasicLevel.DEBUG, "Visit Type: " + node);

        ((Stack) data).push(node);

        logger.log(BasicLevel.DEBUG, "End of Visit Type");
        return null;
    }

    /**
     * qualifiedname could be:
     * - a class field (salary from Employee)
     * - a parameter
     * - a variable
     *
     */
    public Object visit(ASTQualifiedName node, Object data) {
        String name = (String) node.value;
        String[] splitted = splitPath(name);
        Stack stack = (Stack) data;
        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "Visit QualifiedName value=["
                    + node.value + "] stack size=" + stack.size()
                    + " /splitted: [" + buildStringwithout(splitted, -1, ".") + "]");
        }
        try {
			int operatorId = -1;
            if (params != null && params.get(splitted[0]) != null) {
                if (stack.size() > 0 && CONTAINS_PATH_SET.equals(stack.peek())) {
                    logger.log(BasicLevel.DEBUG, "pop: " + stack.pop());
                    String pathset = (String) stack.pop() + "."
                            + NavigatorNodeFactory.GENCLASS_ELEMENT_NAME;
                    logger.log(BasicLevel.DEBUG,
                            "create a member of between gen class element '"
                            + pathset + "' and the parameter '" + name + "'");
                    QueryTreeField qtf =  (QueryTreeField)
                            qt.getTupleStructure().getField(pathset);
                    Expression e = new MemberOf(
                            singletonList(((Object[]) params.get(name))[1]),
                            singletonList(new BasicFieldOperand(qtf)));
                    logger.log(BasicLevel.DEBUG, "push(" + ExpressionPrinter.e2str(e) + ")");
                    stack.push(e);
                } else {
                    logger.log(BasicLevel.DEBUG, "create a parameterOperand with:" + name);
                    stack.push(((Object[]) params.get(name))[1]);
                }
            } else if (vars != null && vars.get(splitted[0]) != null) {
                if (stack.size() > 0 && CONTAINS_PATH_SET.equals(stack.peek())) {
                    logger.log(BasicLevel.DEBUG, "Do not use the pathset of the container operator");
                    stack.pop(); // pop the CONTAINS_PATH_SET
                    stack.pop(); // pop the pathset
                    stack.push(REMOVER);
                } else {
                    logger.log(BasicLevel.DEBUG, "create a variableOperand with:" + name);
                    Field f = qt.getTupleStructure().getField(name);
                    stack.push(new BasicFieldOperand(f));
                }

            } else if (splitted.length >= 2
                    && "isEmpty".equals(splitted[splitted.length - 1])) {
                logger.log(BasicLevel.DEBUG, "Visit IsEmpty: " + node);
                stack.push(REMOVER);

            } else if ((operatorId = isMethodOperator(splitted[splitted.length - 1])) != -1) {
				String theName = curClass + "." + withoutThis(
					buildStringwithout(splitted, splitted.length-1, "."));
				logger.log(BasicLevel.DEBUG, "create a fieldOperand with:" + theName);
				Field f = qt.getTupleStructure().getField(theName);
				Expression e = new BasicFieldOperand(f);
				switch (operatorId) {
				case TO_LOWER_OPERATOR:
					e = new StringLower(e);
					stack.push(e);
					break;
				case TO_UPPER_OPERATOR:
					e = new StringUpper(e);
					stack.push(e);
					break;
				case LENGTH_OPERATOR:
					e = new Length(e);
					stack.push(e);
					break;
				case SUBSTRING_OPERATOR:
					stack.push(e);
					stack.push(STR_OPERAND_SUBSTRING);
					stack.push(splitted[splitted.length - 1]);
					break;
				default:
					stack.push(e);
					stack.push(splitted[splitted.length - 1]);
					break;
				}

            } else if (splitted.length >= 2
                    && "contains".equals(splitted[splitted.length - 1])) {
                String pathset = curClass + "." + withoutThis(
                        buildStringwithout(splitted, splitted.length-1, "."));
                logger.log(BasicLevel.DEBUG, "contains operator: set=" + pathset);
                stack.push(pathset);
                stack.push(CONTAINS_PATH_SET);
            } else {
				String theName = curClass + "." + withoutThis(name);
				logger.log(BasicLevel.DEBUG, "create a fieldOperand with:" + theName);
				Field f = qt.getTupleStructure().getField(theName);
				stack.push(new BasicFieldOperand(f));
            }
        } catch (Exception e) {
            throw new SpeedoQLAbstractVisitor.VisitorException(e);
        }
        return null;
    }
    private static List singletonList(Object element) {
        List res = new ArrayList(1);
        res.add(element);
        return res;
    }
}
