/**
 * Copyright (C) 2001-2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.workingset.lib;

import org.objectweb.perseus.persistence.api.PersistenceException;
import org.objectweb.perseus.persistence.api.TransactionalPersistenceManager;
import org.objectweb.perseus.persistence.api.TransactionalWorkingSet;
import org.objectweb.perseus.persistence.api.WorkingSetLifeCycle;
import org.objectweb.perseus.persistence.api.ConnectionHolder;
import org.objectweb.perseus.persistence.api.State;
import org.objectweb.perseus.persistence.api.VirtualState;
import org.objectweb.perseus.persistence.lib.BasicWorkingSet;
import org.objectweb.speedo.pm.api.ProxyManager;
import org.objectweb.speedo.workingset.api.Transaction;
import org.objectweb.speedo.api.ExceptionHelper;
import org.objectweb.speedo.api.SpeedoProperties;
import org.objectweb.speedo.api.TransactionListener;
import org.objectweb.speedo.mim.api.SpeedoAccessor;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.fractal.api.Component;
import org.objectweb.medor.eval.prefetch.api.PrefetchCache;
import org.objectweb.jorm.api.PMapper;

import javax.jdo.JDOException;
import javax.jdo.PersistenceManager;
import javax.jdo.JDOFatalException;
import javax.jdo.JDOFatalInternalException;
import javax.transaction.Synchronization;
import javax.transaction.Status;
import javax.transaction.TransactionManager;
import javax.naming.InitialContext;
import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;

/**
 *
 * @author S.Chassande-Barrioz
 */
public class SpeedoTransaction
        extends BasicWorkingSet
        implements Transaction, LifeCycleController {

    public final static String PROXY_MANAGER_BINDING = "proxy-manager";
    public final static String MAPPER_BINDING = "mapper";
    public final static String TRANSACTIONAL_PERSISTENCE_MANAGER_BINDING
            = "transactional-persistence-manager";
    public final static String COMPONENT_BINDING = "component";
	public static TransactionListener txListener = null;

    private ProxyManager pm = null;
	private PMapper mapper = null;
    private TransactionalPersistenceManager tpm = null;
    private Transaction thisT;

    private boolean nontransactionalRead;
    private boolean nontransactionalWrite;
    private boolean optimistic;
    private Synchronization synchronization = null;
    private boolean managedEnv = false;
	private List beforeWSPrepare = null;

    /**
     * It activates the working set. This is used to delimit the begining of
     * the working set.
     */
    public void active() throws PersistenceException {
        status = WorkingSetLifeCycle.getNextStatus(
                status, WorkingSetLifeCycle.ACTIVE_ACTION);
    }

	/**
	 * Attaches an entry to the transaction.
	 * @param state the state which must be attached to the transaction
	 */
	public void bind(State state, Object oid) {
		super.bind(state, oid);
		if (beforeWSPrepare != null) {
			logger.log(BasicLevel.DEBUG, "Adding new state in beforeWSPrepare: " + oid);
			int i = 0;
			int size = beforeWSPrepare.size();
			while(i<size && beforeWSPrepare.get(i)!=state) {
				i++;
			}
			if (i==size) {
				beforeWSPrepare.add(state);
			}
		}
	}

	public void beforeWSPrepare() throws PersistenceException {
		try {
			logger.log(BasicLevel.DEBUG, "Starting beforeWSPrepare");
			synchronized(oid2state) {
				beforeWSPrepare = new ArrayList(oid2state.size());
				Iterator it = oid2state.values().iterator();
				while(it.hasNext()) {
					beforeWSPrepare.add(it.next());
				}
			}
			ArrayList exceptions = null;
			// prepare to write (relations consistency management)
			// iterator can not be used (prepareWrite can make
			// new objects dirty, which can modify cacheEntries)
			while(beforeWSPrepare.size() > 0) {
				State state = (State) beforeWSPrepare.remove(0);
				if (state == VirtualState.instance) {
					continue;
				}
				try {
					((SpeedoAccessor) state).prepareWrite();
				} catch (JDOException e) {
					if (exceptions == null) {
						exceptions = new ArrayList();
					}
					exceptions.add(e);
				}
			}

			// close the prefetch buffers associated to the context
			PrefetchCache pc = mapper.getPrefetchCache();
			if (pc != null) {
				pc.invalidatePrefetchBuffer(thisT);
			}
			if (exceptions != null) {
				Exception ie = null;
				if (exceptions.size() == 1) {
					ie = (Exception) exceptions.get(0);
				} else {
					ie = new JDOException(
						"Impossible to prepare the persistent instance to be store",
						(JDOException[]) exceptions.toArray(new JDOException[exceptions.size()]));
				}
				throw new PersistenceException(
					"Impossible to prepare the persistent instance to be store",
					ie);
			}
		} finally {
			beforeWSPrepare = null;
			logger.log(BasicLevel.DEBUG, "Ending beforeWSPrepare");
		}
	}

	public void afterWSPrepare() {
		try {
			logger.log(BasicLevel.DEBUG, "Starting afterWSPrepare");
			synchronized(oid2state) {
				beforeWSPrepare = new ArrayList(oid2state.size());
				Iterator it = oid2state.values().iterator();
				while(it.hasNext()) {
					beforeWSPrepare.add(it.next());
				}
			}
			while(beforeWSPrepare.size() > 0) {
				State state = (State) beforeWSPrepare.remove(0);
				if (state == VirtualState.instance) {
					continue;
				}
				((SpeedoAccessor) state).workingSetClosed();
			}
		} finally {
			beforeWSPrepare = null;
			logger.log(BasicLevel.DEBUG, "Ending afterWSPrepare");
		}
	}

	// IMPLEMENTATION OF THE Transaction INTERFACE //
	//-----------------------------------------------------//

	public void setConnectionHolder(ConnectionHolder ch) {
		connectionHolder = ch;
		connectionHolder.bindWorkingSet(this);
	}

	public JDOFatalException rollBackOnInternalError(Exception _e) {
		Exception ie = ExceptionHelper.getNested(_e);
		if (isActive()) {
			if (managedEnv) {
				logger.log(BasicLevel.INFO, ".");
				String tmName = pm.getPersistenceManagerFactory().getProperties()
						.getProperty(SpeedoProperties.TM_NAME);
				if (tmName == null) {
					return new JDOFatalException("No transaction manager jndi name found in initialisation properties");
				}
				try {
					Object o = new InitialContext().lookup(tmName);
					if (o == null) {
						String msg = "The transaction must be marked as rollbackOnly: JNDI retrieves a null transaction manager for the name '" + tmName + "'.";
						logger.log(BasicLevel.ERROR, msg);
						return new JDOFatalException(msg);
					}
					if (!(o instanceof TransactionManager)) {
						String msg = "The transaction must be marked as rollbackOnly: JNDI retrieves an object which is not a javax.transaction.TransactionManager (JNDI name: " + tmName + "): " + o;
						logger.log(BasicLevel.ERROR, msg);
						return new JDOFatalException(msg);
					}
					javax.transaction.Transaction t = ((TransactionManager) o).getTransaction();
					if (t == null) {
						String msg = "Impossible to rollback an unexisting transaction (TM.getTransaction()=null)";
						logger.log(BasicLevel.ERROR, msg);
						return new JDOFatalException(msg);
					}
					t.setRollbackOnly();
				} catch (Exception e) {
					String msg = "The transaction must be marked as rollbackOnly: Error when lookup the transaction manager in JNDI with the name '"
							+ tmName + "'";
					logger.log(BasicLevel.ERROR, msg, e);
					return new JDOFatalException(msg);
				}
			} else {
				rollback();
			}
			return new JDOFatalException("The current transaction has been rolledback, please retry", ie);
		} else {
			return new JDOFatalException("The current working set occrurs an error, please retry", ie);
		}
	}

    public boolean isManagedEnv() {
        return managedEnv;
    }

    // IMPLEMENTATION OF THE LifeCycleController INTERFACE //
    //-----------------------------------------------------//

    public String getFcState() {
        return null;
    }

    public void startFc() {
        managedEnv = pm.getPersistenceManagerFactory().getProperties()
            .getProperty(SpeedoProperties.MANAGED, "")
            .equals("true");
    }

    public void stopFc() {
    }

    // IMPLEMENTATION OF THE UserBindingController INTERFACE //
    //-------------------------------------------------------//

    public String[] listFc() {
        String[] names = super.listFc();
        String[] itfs = new String[names.length + 3];
        itfs[0] = PROXY_MANAGER_BINDING;
        itfs[1] = TRANSACTIONAL_PERSISTENCE_MANAGER_BINDING;
        itfs[2] = MAPPER_BINDING;
        System.arraycopy(names, 0, itfs, 3, names.length);
        return itfs;
    }

    public Object lookupFc(String c) {
        if (PROXY_MANAGER_BINDING.equals(c))
            return pm;
        else if (TRANSACTIONAL_PERSISTENCE_MANAGER_BINDING.equals(c))
            return tpm;
        else if (MAPPER_BINDING.equals(c))
            return mapper;
        else
            return super.lookupFc(c);
    }

    public void bindFc(String c, Object s) {
        if (PROXY_MANAGER_BINDING.equals(c))
            pm = (ProxyManager) s;
        else if (TRANSACTIONAL_PERSISTENCE_MANAGER_BINDING.equals(c))
            tpm = (TransactionalPersistenceManager) s;
        else if (MAPPER_BINDING.equals(c))
            mapper = (PMapper) s;
        else if (COMPONENT_BINDING.equals(c)) {
            try {
                thisT = (Transaction) ((Component) s).getFcInterface("transaction");
            } catch (Exception e) {
                throw new JDOFatalInternalException(
                        "Impossible to get self transaction",
                        ExceptionHelper.getNested(e));
            }
        } else
            super.bindFc(c, s);
    }

    public void unbindFc(String c) {
        if (PROXY_MANAGER_BINDING.equals(c))
            pm = null;
        else if (TRANSACTIONAL_PERSISTENCE_MANAGER_BINDING.equals(c))
            tpm = null;
        else if (MAPPER_BINDING.equals(c))
            mapper = null;
        else
            super.unbindFc(c);
    }

	public void setStatus(byte status) throws PersistenceException {
		try {
			if (status == TransactionalWorkingSet.CTX_PREPARED) {
				beforeWSPrepare();
			} else if (status ==TransactionalWorkingSet.CTX_PREPARED_OK) {
				afterWSPrepare();
			}
		} finally {
			super.setStatus(status);
		}
	}

    // IMPLEMENTATION OF THE javax.jdo.Transactionn INTERFACE //
    //--------------------------------------------------------//

    public void begin() {
		logger.log(BasicLevel.DEBUG, "Begin the transaction");
		if (txListener != null) {
			txListener.transactionBegun(this);
		}
        try {
            tpm.begin(thisT);
        } catch (PersistenceException e) {
            Exception ie = ExceptionHelper.getNested(e);
            logger.log(BasicLevel.ERROR,
                    "Error during the begin of the transaction:", ie);
            throw new JDOException("", ie);
        }
    }

    public void commit() {
        logger.log(BasicLevel.DEBUG, "Commit the transaction, working set size: " + oid2state.size());
        if (synchronization != null)
            synchronization.beforeCompletion();
		//register working set size for statistics
		int size = oid2state.size();
        boolean validated = false;
        try {
            tpm.commit(thisT);
			validated = true;
        } catch (PersistenceException e) {
			Exception ie = ExceptionHelper.getNested(e);
			if (ie instanceof JDOException) {
				throw (JDOException) ie;
			} else {
            	throw new JDOFatalException(
					"Rollback due to an exception at the commit time: ", ie);
			}
        } finally {
			if (synchronization != null)
				synchronization.afterCompletion((validated
						? Status.STATUS_COMMITTED : Status.STATUS_ROLLEDBACK));
			if (txListener != null) {
				if (validated) {
					txListener.transactionCommitted(this, size);
				} else {
					txListener.transactionAborted(this, size);
				}
			}
        }
    }

    public void rollback() {
        logger.log(BasicLevel.INFO, "Roll back a transaction: ");
		int size = oid2state.size();
        try {
            tpm.rollback(thisT);
        } catch (PersistenceException e) {
            Exception ie = ExceptionHelper.getNested(e);
            logger.log(BasicLevel.ERROR,
                    "Error during the rollback of the transaction:", ie);
            throw new JDOException("", ie);
        } finally {
			if (txListener != null) {
				txListener.transactionAborted(this, size);
			}
			if (synchronization != null)
				synchronization.afterCompletion(Status.STATUS_ROLLEDBACK);
		}
    }

    public boolean isActive() {
        switch(status){
        case TransactionalWorkingSet.CTX_ACTIVE_TRANSACTIONAL:
        case TransactionalWorkingSet.CTX_PREPARED:
        case TransactionalWorkingSet.CTX_PREPARED_OK:
        case TransactionalWorkingSet.CTX_PREPARED_FAIL:
            return true;
        case TransactionalWorkingSet.CTX_ACTIVE:
        case TransactionalWorkingSet.CTX_COMMITTED:
        case TransactionalWorkingSet.CTX_ABORTED:
        case TransactionalWorkingSet.CTX_CLOSED:
        default:
            return false;
        }
    }

    public void setNontransactionalRead(boolean b) {
        nontransactionalRead = b;
    }

    public boolean getNontransactionalRead() {
        return nontransactionalRead;
    }

    public void setNontransactionalWrite(boolean b) {
        nontransactionalWrite = b;
    }

    public boolean getNontransactionalWrite() {
        return nontransactionalWrite;
    }

    public void setRetainValues(boolean b) {
        setWSRetainValues(b);
    }

    public boolean getRetainValues() {
        return getWSRetainValues();
    }

    public void setRestoreValues(boolean b) {
        setWSRestoreValues(b);
    }

    public boolean getRestoreValues() {
        return getWSRestoreValues();
    }

    public void setOptimistic(boolean b) {
        optimistic = b;
    }

    public boolean getOptimistic() {
        return optimistic;
    }

    public void setSynchronization(Synchronization s) {
        synchronization = s;
    }

    public Synchronization getSynchronization() {
        return synchronization;
    }

    public PersistenceManager getPersistenceManager() {
        return pm;
    }
}
