/**
 * Copyright (C) 2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.runtime.concurrency;

import org.objectweb.speedo.SpeedoTestHelper;
import org.objectweb.speedo.api.ExceptionHelper;
import org.objectweb.speedo.pobjects.userid.BasicB;
import org.objectweb.speedo.pobjects.basic.BasicA;
import org.objectweb.util.monolog.api.BasicLevel;

import javax.jdo.PersistenceManager;
import javax.jdo.JDOException;
import javax.jdo.JDOFatalException;

import junit.framework.Assert;

import java.util.ArrayList;
import java.util.Vector;
import java.util.Iterator;

/**
 * Tests the high concurrency of the Speedo (more than 100 users)
 *
 * @author S.Chassande-Barrioz
 */
public class TestManyUsers extends SpeedoTestHelper {

	public static Vector rollbackExceptions = new Vector();
	public static Vector errors = new Vector();


	public TestManyUsers(String s) {
		super(s);
	}

	protected String getLoggerName() {
		return LOG_NAME + ".rt.concurrency.TestManyUsers";
	}

	/**
	 * Tests the concurrency of the initialization of a class (mapping)
	 * (1000 users)
	 */
	public void testLoading() {
		final int nbThread = Integer.getInteger(
				"conform.org.objectweb.speedo.rt.concurrency.TestManyUsers.testLoading.thread",
				1000).intValue();
		logger.log(BasicLevel.INFO, "testLoading.nbThread: " + nbThread);
		Thread[] ts = new Thread[nbThread];
		for (int i = 0; i < nbThread; i++) {
			ts[i] = new Thread(
					new Runnable() {
						public void run() {
							try {
								PersistenceManager pm = pmf.getPersistenceManager();
								Thread.sleep(1);
								pm.getObjectIdClass(BasicB.class);
								pm.close();
								Assert.assertTrue(true);
							} catch (Exception e) {
								e.printStackTrace();
								fail(e.getMessage());
							}
						}
					}
			);
		}
		for (int i = 0; i < nbThread; i++) {
			ts[i].start();
		}
		try {
			for (int i = 0; i < nbThread; i++) {
				ts[i].join();
			}
		} catch (InterruptedException e) {
			fail(e.getMessage());
		}
	}

	/**
	 * Tests the concurrency of several writer on a same persistent object.
	 * (500 users)
	 */
	public void testBAccess() {
		final int nbThread = Integer.getInteger(
				"conform.org.objectweb.speedo.rt.concurrency.TestManyUsers.testBAccess.thread",
				500).intValue();
		logger.log(BasicLevel.INFO, "testBAccess.nbThread: " + nbThread);
		final int initialValue = 0;
		rollbackExceptions.clear();
		errors.clear();
		Thread[] ts = new Thread[nbThread];
		BasicA ba = new BasicA();
		ba.writeF1("1");
		ba.writeF2(initialValue);
		PersistenceManager pm = pmf.getPersistenceManager();
		pm.currentTransaction().begin();
		pm.makePersistent(ba);
		final Object id = pm.getObjectId(ba);
		Assert.assertNotNull("Null object identifier", id);
		pm.currentTransaction().commit();
		pm.close();
		for (int i = 0; i < nbThread; i++) {
			final int _i = i;
			ts[i] = new Thread(
					new Runnable() {
						public void run() {
							PersistenceManager pm = pmf.getPersistenceManager();
							logger.log(BasicLevel.DEBUG, _i + " got a PM");
							boolean rollback = false;
							try {
								pm.currentTransaction().begin();
								logger.log(BasicLevel.DEBUG, _i + " begin tx");
								BasicA a = (BasicA) pm.getObjectById(id, false);
								a.writeF2(a.readF2() + 1);
								pm.currentTransaction().commit();
								logger.log(BasicLevel.DEBUG, _i + " finished");
							} catch (JDOFatalException e) {
								rollback = true;
								rollbackExceptions.add(e);
								logger.log(BasicLevel.DEBUG, "Tx " + _i + " has been rolledback");
							} catch (Exception e) {
								Exception ie = ExceptionHelper.getNested(e);
								errors.add(ie);
								logger.log(BasicLevel.ERROR, _i + " has a problem", ie);
								pm.currentTransaction().rollback();
							} finally {
								try {
									pm.close();
								} catch (JDOException e) {
									logger.log(BasicLevel.ERROR, "tx " + _i
																 + " has been "
																 + (rollback ? "rolledback" : "committed")
																 + " and the close occurs an error", e);
									throw e;
								}
							}
						}
					}
			);
		}
		for (int i = 0; i < nbThread; i++) {
			logger.log(BasicLevel.DEBUG, "Start " + i);
			ts[i].start();
		}
		int val;
		try {
			ArrayList al = new ArrayList(nbThread);
			for (int i = 0; i < nbThread; i++) {
				logger.log(BasicLevel.DEBUG, "Join " + i);
				ts[i].join(2000);
				if (ts[i].isAlive()) {
					al.add(new Integer(i));
					logger.log(BasicLevel.DEBUG, i + " is not finished after" +
												 " the delay, it could be blocked");
				}
			}
			//logger.log(BasicLevel.INFO, getDG(pmf));
			if (al.size() > 0) {
				for (Iterator it = al.iterator(); it.hasNext();) {
					int th = ((Integer) it.next()).intValue();
					if (!ts[th].isAlive()) {
						logger.log(BasicLevel.DEBUG, th + " is late but ok.");
						it.remove();
					}
				}
				if (al.size() > 0) {
					fail("Thread " + al + " blocked!");
				}
			}
		} catch (InterruptedException e) {
			//fail(e.getMessage());
		} finally {
			pm = pmf.getPersistenceManager();
			ba = (BasicA) pm.getObjectById(id, false);
			val = ba.f2;
			pm.currentTransaction().begin();
			pm.deletePersistent(ba);
			pm.currentTransaction().commit();
			pm.close();
			logger.log(BasicLevel.INFO, "Commited transaction rate: "
										+ (100 - (((errors.size() + rollbackExceptions.size()) * 100) / nbThread)) + "%");
		}
		if (errors.size() > 0) {
			fail("There are " + errors.size() + "/" + nbThread + " errors during the run");
		}
		Assert.assertEquals("Bad f2 value",
							nbThread + initialValue - rollbackExceptions.size(), val);
	}

	/**
	 * Tests the concurrency of several writer on a same persistent object.
	 * (500 users)
	 */
	public void testCAccess() {
		final int nbThread = Integer.getInteger(
				"conform.org.objectweb.speedo.rt.concurrency.TestManyUsers.testBAccess.thread",
				500).intValue();
		logger.log(BasicLevel.INFO, "testBAccess.nbThread: " + nbThread);
		final int initialValue = 0;
		rollbackExceptions.clear();
		errors.clear();
		Thread[] ts = new Thread[nbThread];
		BasicA ba = new BasicA();
		ba.writeF1("1");
		ba.writeF2(initialValue);
		PersistenceManager pm = pmf.getPersistenceManager();
		pm.currentTransaction().begin();
		pm.makePersistent(ba);
		final BasicA a = ba;
		Object id = pm.getObjectId(ba);
		pm.currentTransaction().commit();
		pm.close();
		for (int i = 0; i < nbThread; i++) {
			final int _i = i;
			ts[i] = new Thread(
					new Runnable() {
						public void run() {
							PersistenceManager pm = pmf.getPersistenceManager();
							logger.log(BasicLevel.DEBUG, _i + " got a PM");
							boolean rollback = false;
							try {
								pm.currentTransaction().begin();
								logger.log(BasicLevel.DEBUG, _i + " begin tx");
								a.incF2();
								pm.currentTransaction().commit();
								logger.log(BasicLevel.DEBUG, _i + " finished");
							} catch (JDOFatalException e) {
								rollback = true;
								rollbackExceptions.add(e);
								logger.log(BasicLevel.DEBUG, "Tx " + _i + " has been rolledback");
							} catch (Exception e) {
								Exception ie = ExceptionHelper.getNested(e);
								errors.add(ie);
								logger.log(BasicLevel.ERROR, _i + " has a problem", ie);
								pm.currentTransaction().rollback();
							} finally {
								try {
									pm.close();
								} catch (JDOException e) {
									logger.log(BasicLevel.ERROR, "tx " + _i
																 + " has been "
																 + (rollback ? "rolledback" : "committed")
																 + " and the close occurs an error", e);
									throw e;
								}
							}
						}
					}
			);
		}
		for (int i = 0; i < nbThread; i++) {
			logger.log(BasicLevel.DEBUG, "Start " + i);
			ts[i].start();
		}
		int val;
		try {
			ArrayList al = new ArrayList(nbThread);
			for (int i = 0; i < nbThread; i++) {
				logger.log(BasicLevel.DEBUG, "Join " + i);
				ts[i].join(2000);
				if (ts[i].isAlive()) {
					al.add(new Integer(i));
					logger.log(BasicLevel.DEBUG, i + " is not finished after" +
												 " the delay, it could be blocked");
				}
			}
			//logger.log(BasicLevel.INFO, getDG(pmf));
			if (al.size() > 0) {
				for (Iterator it = al.iterator(); it.hasNext();) {
					int th = ((Integer) it.next()).intValue();
					if (!ts[th].isAlive()) {
						logger.log(BasicLevel.DEBUG, th + " is late but ok.");
						it.remove();
					}
				}
				if (al.size() > 0) {
					fail("Thread " + al + " blocked!");
				}
			}
		} catch (InterruptedException e) {
			//fail(e.getMessage());
		} finally {
			pm = pmf.getPersistenceManager();
			ba = (BasicA) pm.getObjectById(id, false);
			val = ba.f2;
			pm.currentTransaction().begin();
			pm.deletePersistent(ba);
			pm.currentTransaction().commit();
			pm.close();
			logger.log(BasicLevel.INFO, "Commited transaction rate: "
										+ (100 - (((errors.size() + rollbackExceptions.size()) * 100) / nbThread)) + "%");
		}
		if (errors.size() > 0) {
			fail("There are " + errors.size() + "/" + nbThread + " errors during the run");
		}
		Assert.assertEquals("Bad f2 value",
							nbThread + initialValue - rollbackExceptions.size(), val);
	}

/*
	private static String getDG(PersistenceManagerFactory pmf) {
		Map m = null; //RWFifoLock.dg.successors;
		StringBuffer sb = new StringBuffer("dependency Graph: ");
		if (m.size() > 0) {
			List waiters = new ArrayList(m.keySet());
			Set s = new HashSet(waiters.size() * 2);
			s.addAll(m.keySet());
			s.addAll(m.values());
			List all = new ArrayList(s);
			for (int i = 0; i < waiters.size(); i++) {
				sb.append("\nws");
				sb.append(all.indexOf(waiters.get(i)));
				sb.append(" = > ws");
				sb.append(all.indexOf(m.get(waiters.get(i))));
			}
		}
		return sb.toString();
	}
*/
}
